package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.util.*;
import polyglot.types.*;
import polyglot.ext.predj.types.PredJTypeSystem;
import polyglot.ext.predj.types.PredJPredMethodInstance;
import polyglot.ext.predj.types.Predicate;
import polyglot.visit.*;
import polyglot.ext.predj.util.*;

import polyglot.ext.jl5.types.FlagAnnotations;
import polyglot.ext.jl5.visit.JL5AmbiguityRemover;

import java.util.*;

/** Adds a predicate to method declarations. **/
public class PredJPredMethodDecl_c extends PredJAbstractMethodDecl_c
    implements PredJPredMethodDecl {

    protected PredicateNode predNode;

	// keeps track of the specialized types of formal parameters in
	// a predicate; a set of SpecializerNodes
    protected Set formalBindings;

	// keeps track of the new variable bindings in a predicate; a
	// list of BindingSpecializerNodes; it's a list so we can keep
	// the order of vars, since later bindings may refer to
	// earlier ones
    protected List localBindings;
    
    public PredJPredMethodDecl_c(Position pos, FlagAnnotations flags,
				 TypeNode returnType, String name,
				 List formals,
				 List throwTypes, Block body,
				 List typeParams, PredicateNode predNode) {
	super(pos, flags, returnType, name, formals, throwTypes, body,
	      typeParams);
	    //assert(predNode != null);
	this.predNode = predNode;
    }

	/** Get the predicate of this method. */
    public PredicateNode predNode() {
	return this.predNode;
    }

	// does this method have a predicate?
    public boolean hasPredicate() {
	return true;
    }
    
    public Set formalBindings() {
	return formalBindings;
    }

    public List localBindings() {
	return localBindings;
    }

    public PredJPredMethodDecl bindings(Set formalBindings,
					List localBindings) {
	PredJPredMethodDecl_c n = (PredJPredMethodDecl_c) copy();
	n.formalBindings = formalBindings;
	n.localBindings = localBindings;
	return n;
    }

        /** Build type objects for the method. */
    public NodeVisitor disambiguateEnter(AmbiguityRemover ar)
	throws SemanticException {
	HaltingVisitor hv = (HaltingVisitor) super.disambiguateEnter(ar);
	    /* skip the predicate when disambiguating signatures;
	       the predicate can only be disambiguated properly once
	       all the class's members have been added (in the addMembers
	       pass).  otherwise, we get spurious compile-time errors
	       on things like field references. */
	AmbiguityRemover.Kind k = ar.kind();
        if (k == AmbiguityRemover.SIGNATURES ||
	    k == JL5AmbiguityRemover.TYPE_VARS) {
	    return hv.bypass(predNode);
	} else {
	    return hv;
	}
    }
    
    public MethodInstance makeMethodInstance(ClassType ct, TypeSystem ts)
	throws SemanticException {

	MethodInstance mi = super.makeMethodInstance(ct, ts);
	
	PredJTypeSystem pts = (PredJTypeSystem) ts;

	return makePredMethodInstance(mi, pts);
    }

    protected static PredJPredMethodInstance makePredMethodInstance(
	MethodInstance mi, PredJTypeSystem pts) {
	Predicate pred;
	return pts.predMethodInstance(
	    mi.position(), mi.container(), mi.flags(), mi.returnType(),
	    mi.name(), mi.formalTypes(), mi.throwTypes(),
	    pts.unknownPredicate());
    }	
    
	
    /** Reconstruct the method. */
    protected MethodDecl_c reconstruct(TypeNode returnType, List formals,
				       PredicateNode predNode,
				       List throwTypes, Block body) {
	if (returnType != this.returnType ||
	    ! CollectionUtil.equals(formals, this.formals) ||
	    predNode != this.predNode ||
	    ! CollectionUtil.equals(throwTypes, this.throwTypes) ||
	    body != this.body) {
	    PredJPredMethodDecl_c n = (PredJPredMethodDecl_c) copy();
	    n.returnType = returnType;
	    n.formals = TypedList.copyAndCheck(formals, Formal.class, true);
	    n.predNode = predNode;
	    n.throwTypes = TypedList.copyAndCheck(throwTypes,
						  TypeNode.class, true);
	    n.body = body;
	    return n;
	}

	return this;
    }

      /** Visit the children of the method. */
    public Node visitChildren(NodeVisitor v) {
	TypeNode returnType = (TypeNode) visitChild(this.returnType, v);
	List formals = visitList(this.formals, v);
	PredicateNode predNode = (PredicateNode) visitChild(this.predNode, v);
	List throwTypes = visitList(this.throwTypes, v);
	Block body = (Block) visitChild(this.body, v);
	return reconstruct(returnType, formals, predNode, throwTypes, body);
    }

      public Node buildTypes(TypeBuilder tb) throws SemanticException {
	  
	  MethodInstance mi =
	      ((MethodDecl) super.buildTypes(tb)).methodInstance();
	  
	  PredJTypeSystem ts = (PredJTypeSystem) tb.typeSystem();

	  PredJPredMethodInstance pmi =
	      makePredMethodInstance(mi, ts);
	  return methodInstance(pmi);
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
	PredJPredMethodDecl n = (PredJPredMethodDecl) super.typeCheck(tc);

	if (n.flags().isAbstract()) {
	    throw new SemanticException(
		"An abstract method cannot have a predicate.", position());
	}
	return n;
    }

    public Node createPredicate(Context c, PredJTypeSystem ts) {
	PredJPredMethodDecl n = this;
	
	    // create the type information for the method's predicate
	PredJPredMethodInstance oldMI =
	    (PredJPredMethodInstance) n.methodInstance();
	PredJPredMethodInstance newMI =
	    oldMI.predicate(n.predNode().toPredicate(ts));

	    // replace the old mi with the new one in the associated
	    // class type object
	List mis = c.currentClass().methods();
	Iterator it = mis.iterator();
	while(it.hasNext()) {
	    MethodInstance mi = (MethodInstance) it.next();
	    if (mi.equals(oldMI)) {
		it.remove();
		break;
	    }
	}
	mis.add(newMI);
	    
	return n.methodInstance(newMI);
    }
    
}
