package polyglot.ext.predj.types;

import polyglot.types.*;
import polyglot.ast.Binary;
import polyglot.ast.Unary;
import polyglot.ast.Special;
import polyglot.ext.jl.types.*;

import java.io.*;
import java.util.*;

// an interface for interacting with CVC Lite

public class CVCLInterface_c implements TheoremProverInterface {

    protected PredJTypeSystem ts;

	// the stream for writing axioms and queries to CVC Lite
    protected BufferedWriter bw;
	// the stream for reading results from CVC Lite
    protected BufferedReader br;

    protected final boolean DEBUG = true;
    protected FileWriter fw=null;
    
    protected Set seenSpecializers = new HashSet();
    protected Set seenTargets = new HashSet();

    protected static String cvc_executable;

    protected static String cvc_options = "+interactive";
    
	// used to separate names on a path
    protected String CVC_PathSep() { return "$"; }

    protected String uniquify(String name) {
	return CVC_PathSep() + name + CVC_PathSep();
    }

	// the type of classes
    protected static String CVC_Class = "Class";
    protected String CVC_Class() { return uniquify(CVC_Class); }
	// the type of instances (objects)
    protected static String CVC_Object = "Object";
    protected String CVC_Object() { return uniquify(CVC_Object); }
	// the instanceof predicate
    protected static String CVC_Instanceof = "instanceof";
	// the null value
    protected static String CVC_Null = "null";
	// 'this'; it must be defined in a manner consistent with how
	// 'this' is translated in full paths, e.g this.f
    protected static String CVC_This = "this";

	// used for array syntax (cvc doesn't like the [] brackets)
    protected static String CVC_ArraySyntax = "Array";
    protected String CVC_ArraySyntax() { return uniquify(CVC_ArraySyntax); }

    protected static String CVC_true = "TRUE";
    protected static String CVC_real = "REAL";
    protected static String CVC_bool = "BOOLEAN";
    protected static String CVC_AND = "AND";
    protected static String CVC_OR = "OR";
    protected static String CVC_NOT = "NOT";
    protected static String CVC_IMP = "=>";
    protected static String CVC_IFF = "<=>";
    protected static String CVC_XOR = "XOR";
    protected static String CVC_GT = ">";
    protected static String CVC_GE = ">=";
    protected static String CVC_LT = "<";
    protected static String CVC_LE = "<=";
    protected static String CVC_EQ = "=";
    protected static String CVC_NE = "/=";
    protected static String CVC_ADD = "+";
    protected static String CVC_SUB = "-";
    protected static String CVC_MUL = "*";
    protected static String CVC_NEG = "-";
    protected static String CVC_array = "ARRAY REAL OF ";

    public CVCLInterface_c(PredJTypeSystem ts) {
	this.ts = ts;
	determineCVCExecutable();
	try {
	    // start the CVC Lite process, and create its output and input
	    // streams
	    Runtime rt = Runtime.getRuntime();
	    String[] command = {cvc_executable, cvc_options};
	    Process p = rt.exec(command);
	    OutputStream os = p.getOutputStream();
	    OutputStreamWriter osw = new OutputStreamWriter(os);
	    bw = new BufferedWriter(osw);
	    InputStream is = p.getInputStream();
	    InputStreamReader isr = new InputStreamReader(is);
	    br = new BufferedReader(isr);

	    // give CVC some basic declarations, used in all queries
	    StringBuffer sb = new StringBuffer();
	    writePredicateIndependentDecls(sb);	
	    bw.write(sb.toString() + '\n');
            if (DEBUG) {
                fw = new FileWriter(".predj-cvc", false);
	        debugWrite(sb.toString() + '\n');
            }
	} catch(IOException e) {
	    throw new RuntimeException("Problem with " +
				       executableName() + " initialization");
	}
    }
    
	// compute the partial order among a bunch of predicates.
	// this is currently too stupid to be useful, so we're not using it.
	// in particular, axioms about the class hierarchy should be generated
	// per query, rather than all at once, so that we're not taking the
	// cross product of a huge number of specializers and targets.
//     public void partialOrder(Predicate[] preds, List[] po) {
// 	try {
// 	    StringBuffer sb = new StringBuffer();
// 	    writePredicateIndependentDecls(sb);

// 	    String[] predStrs = new String[preds.length];
// 	    predStrs[0] = pred2String(preds[0]);
// 	    Set specializers = preds[0].specializers();
// 	    Set targets = preds[0].targets();
// 	    for (int i = 1; i < preds.length; i++) {
// 		predStrs[i] = pred2String(preds[i]);
// 		specializers.addAll(preds[i].specializers());
// 		targets.addAll(preds[i].targets());
// 	    }
// 	    declareSpecializers(specializers, sb);
// 	    declareTargets(targets, sb);

// 		// assert axioms about null
// 	    assertNullSemantics(specializers, sb);

// 		// add axioms about the relationship among the specializers
// 	    assertClassHierarchy(specializers, targets, sb);

// 		// add axioms about integers, since CVC only includes
// 		// the theory of rationals.
// 		//assertIntegerAxioms(sb);

// 		// compute the partial order
// 	    for(int i=1; i < preds.length; i++) {
// 		for (int j=0; j < i; j++) {
// 		    writeQuery(implies(predStrs[i], predStrs[j]), sb);
// 		    writeQuery(implies(predStrs[j], predStrs[i]), sb);
// 		}
// 	    }
// 	    List output = cvcQuery(sb);

// 		// interpret the results, populating the partial order table
// 	    int count = 0;
// 	    po[0] = new LinkedList();
// 	    for (int i=1; i < preds.length; i++) {
// 		po[i] = new LinkedList();
// 		for (int j=0; j < i; j++) {
// 		    if (interpretCVCOutput((String) output.get(count))) {
// 			po[i].add(new Integer(j));
// 		    }
// 		    count++;
// 		    if (interpretCVCOutput((String) output.get(count))) {
// 			po[j].add(new Integer(i));
// 		    }
// 		    count++;
// 		}
// 	    }
// 	} catch(IOException e) {
// 	    throw new RuntimeException("Problem with CVCL " +
// 				       "partial order testing");
// 	}
//     }

	// does m1 logically imply m2?
    public boolean implies(Predicate p1, Predicate p2) {
	Predicate p =
	    ts.createPredicateBinary(
		ts.createPredicateUnary(Unary.NOT, p1, p1.type()),
		Binary.COND_OR, p2, ts.Boolean());
	return isTautology(p);
// 	try {
// 	    StringBuffer sb = new StringBuffer();
// 	    Set specializers = p1.specializers();
// 	    specializers.addAll(p2.specializers());
// 	    declareSpecializers(specializers, sb);

// 	    Set targets = p1.targets();	    
// 	    targets.addAll(p2.targets());

// 	    Set names = p1.names();
// 	    names.addAll(p2.names());
// 	    declareTargets(names, sb);

// 	    assertAxioms(specializers, targets, sb);

// 	    writeQuery(implies(pred2String(p1), pred2String(p2)), sb);
	    
// 	    boolean result = cvcQuery(sb);
// //	    boolean result = interpretCVCOutput(output);

// //	    System.out.println("result is " + result);
// 	    return result;

// 	} catch(IOException e) {
// 	    throw new RuntimeException("Problem with CVCL " +
// 				       "implication testing");
// 	}
    }

	// is the given predicate logically true?
    public boolean isTautology(Predicate p) {
	try {
	    StringBuffer sb = new StringBuffer();	    
	    
	    Set specializers = p.specializers();
	    declareSpecializers(specializers, sb);	    		
	    
	    declareTargets(p.names(), sb);

	    Set targets = p.targets();
	    Iterator i = targets.iterator();
		// Since a PredicateArrayAccess is created recursively,
		// for a multidimensional array access, more than one
		// PredicateArrayAccess is created (one more for each of the underlying
		// accesses).  These underlying accesses are not needed for
		// CVC axiom generation/queries (actually, they cause problems with CVC
		// reasoning!).  Below, any PredicateArrayAccess that has type
		// of array (meaning that the access results in an array type),
		// is removed from the targets() list to prevent the CVC problems.		
	    while(i.hasNext()) {
		Predicate pred = (Predicate)i.next();
		if (pred instanceof PredicateArrayAccess
		    && ((PredicateArrayAccess)pred).type().isArray()) {
		    i.remove();
		}
		    
	    }		
	    
	    assertAxioms(specializers, targets, sb);
	    
	    writeQuery(pred2String(p), sb);	    
	    
	    boolean result = cvcQuery(sb);
//	    boolean result = interpretCVCOutput(output);
	    return result;
	    
	} catch(IOException e) {
	    throw new RuntimeException("Problem with CVCL " +
		"tautology testing");
	}
    }
	    

	    // declare the types and function symbols of our predicate
	    // language, which are independent of any particular predicate
    protected void writePredicateIndependentDecls(StringBuffer sb)
	throws IOException {
	writeCommand(typeDecl(CVC_Class()), sb);
	writeCommand(typeDecl(CVC_Object()), sb);
	String[] instanceofArgs = {CVC_Object(), CVC_Class()};
	writeCommand(predDecl(CVC_Instanceof, instanceofArgs), sb);
	    // declare null and this
	declareVarOfType(CVC_Null, CVC_Object(), sb);
	declareVarOfType(CVC_This, CVC_Object(), sb);
	seenTargets.add(CVC_This);
	newLine(sb);
    }


    protected void declareSpecializers(Set specializers, StringBuffer sb)
	throws IOException {
	Iterator i = specializers.iterator();
	while (i.hasNext()) {
	    Type t = (Type) i.next();
	    String s = ts.erasure(t).toString();
	    if (!seenSpecializers.contains(s)) {
		seenSpecializers.add(s);
		declareVarOfType(s, CVC_Class(), sb);
	    }
	}
    }

    protected String mangleArrayName(String name, ArrayType type) {
	int dims = type.dims();
	
	if (type.ultimateBase().isNumeric())
	    name += uniquify(CVC_real);
	else
	    name += CVC_Object();

	for(int i=0; i < dims; i++) 
		    name += CVC_ArraySyntax();

	return name;
    }
    
    protected void declareTargets(Set targets, StringBuffer sb)
	throws IOException {
	Iterator iter = targets.iterator();

	while (iter.hasNext()) {		
	    PredicateTarget target = (PredicateTarget) iter.next();  
	    String s = target.name();
	    
	    if (target.type().isArray()) {				
		String s2 = mangleArrayName(s, target.type().toArray());
		
		if (!seenTargets.contains(s2)) {
		    seenTargets.add(s2);		    		    
		    declareVarOfType(s2, type2String(target.type()), sb);
		}
		if (!seenTargets.contains(s)) {
		    seenTargets.add(s);		    		    
		    declareVarOfType(s, CVC_Object(), sb);
		}
	    }
	    if (!seenTargets.contains(s)) {
		seenTargets.add(s);
                if (target instanceof PredicateCall) {
                    PredicateCall pc = (PredicateCall) target;
                    List pcArgTypes = pc.argumentTypes();
                    String[] argTypes = new String[pcArgTypes.size()];
                    for (int i = 0; i < argTypes.length; i++)
                        argTypes[i] = type2String((Type)pcArgTypes.get(i));
                    declareFunOfType(s, argTypes, type2String(pc.type()), sb);
                }
                else
	    	    declareVarOfType(s, type2String(target.type()), sb);
	    }		
	}
	
	newLine(sb);
    }
	
    protected void declareVarsOfType(Set vars, String type,
				     StringBuffer sb)
	throws IOException {
	Iterator i = vars.iterator();
	while(i.hasNext()) {
	    declareVarOfType(i.next().toString(), type, sb);
	}
	newLine(sb);
    }

    protected void declareVarOfType(String var, String type,
				    StringBuffer sb) throws IOException {
	writeCommand(decl(cleanIdentifier(var), type), sb);
    }

    protected void declareFunOfType(String var, String[] argTypes,
                                    String retType, StringBuffer sb)
                                                            throws IOException {
        writeCommand(funDecl(cleanIdentifier(var), argTypes, retType), sb);
    }

    protected void assertAxioms(Set specializers, Set targets,
				StringBuffer sb) throws IOException {
		// assert axioms about null
	    assertNullSemantics(specializers, sb);

	    Set refTargets = referenceTargets(targets);

		// assert that each reference target is either null or
		// is an instance of its static type
	    assertTargetTypes(refTargets, specializers, sb);
	    
		// add axioms about the relationship among the specializers
	    assertClassHierarchy(specializers, refTargets, sb);
		// add axioms about the relationship between a target
		// and its fields
	    assertFieldSemantics(targets, sb);
    }

	// assert that null is not an instance of any class
    protected void assertNullSemantics(Set specializers, StringBuffer sb)
	throws IOException {
	Iterator iter = specializers.iterator();
	while (iter.hasNext()) {
	    String cls = cleanIdentifier(iter.next().toString());
	    String assertion = assertion(not(instanceOf(CVC_Null, cls)));
	    writeCommand(assertion, sb);
	}
//	newLine(sb);
//	writeCommand(assertion(binaryOp(CVC_This, CVC_Null, CVC_NE)), sb);
//	newLine(sb);
    }

	// assert that each reference target is either null or
	// is an instance of its static type
    protected void assertTargetTypes(Set refTargets, Set specializers,
				     StringBuffer sb)
	throws IOException {
	Iterator i = refTargets.iterator();
	
	while (i.hasNext()) {
		//PredicateTarget t = (PredicateTarget) i.next();
	    Predicate t = (Predicate) i.next();
	    Type ty = t.type();
		// as an optimization, only assert the axiom if the
		// static type is in our list of specializers.  if
		// it's left off, that means the axiom is unnecessary.
		// this is important because these axioms, which use
		// disjunction, make the theorem prover significantly
		// slower.
	    if (specializers.contains(ty)) {
		String name = pred2String(t);		    
		String instof = instanceOf(name,
					   cleanIdentifier(ty.toString()));
		String assertion;
		    // "this" can never be null
		if (name.equals(CVC_This)) {
		    assertion = assertion(instof);
		} else {
		    assertion = assertion(
			or(binaryOp(name, CVC_Null, CVC_EQ), instof));
		}
		writeCommand(assertion, sb);
	    }
	}	
    }
    
	// assert axioms that characterize the relevant portion of the
	// class hierarchy
    protected void assertClassHierarchy(Set specializers,
					Set refTargets, StringBuffer sb)
	throws IOException {
	Type[] classes = (Type[]) specializers.toArray(new Type[0]);

	    // now assert pairwise subtype and disjointness axioms
	    // among the specializers, exploded onto all the reference
	    // targets
	boolean b1, b2;	    
	for (int i = 0; i < classes.length; i++) {
	    for (int j=0; j < i; j++) {
		    // isSubtye changed to isImplicitCastValid for
		    // more generality (namely, over array types)
		b1 = ts.isImplicitCastValid(classes[i], classes[j]);
		b2 = ts.isImplicitCastValid(classes[j], classes[i]);		
		
		if (b1) {
		    assertSubtype(classes[i], classes[j], refTargets, sb);	   		
		}
		else if (b2) {		    
		    assertSubtype(classes[j], classes[i], refTargets, sb);
		}
		else {		    
		    Type classI = classes[i];
		    Type classJ = classes[j];
		    
			// If we have arrays, strip the dimensions off.
			// If the dimensions aren't the same, they're disjoint.			
		    while (classI.isArray() && classJ.isArray()) {
			classI = classI.toArray().base();
			classJ = classJ.toArray().base();
		    }		    
		    if ( classI.isArray() || classJ.isArray()
			 || !(classI.isClass() && classJ.isClass()) // Might be primitive array.
			 || !(classI.toClass().flags().isInterface()
			      || classJ.toClass().flags().isInterface()) )
			assertDisjoint(classes[i], classes[j], refTargets, sb);			
		}
	    }
	}
	newLine(sb);
    }

    


    protected void assertFieldSemantics(Set targets, StringBuffer sb)
	throws IOException {
	String[] targetNames = new String[targets.size()];
	Iterator iter = targets.iterator();
	int count = 0;
	while (iter.hasNext()) {
		//PredicateTarget target = (PredicateTarget) iter.next();
	    Predicate target = (Predicate) iter.next();
		//targetNames[count] = cleanIdentifier(target.name());
	    targetNames[count] = pred2String(target);
	    count++;
	}

	    // map suffixes to targets
	Map m = new HashMap();
	for (int i = 0; i < targetNames.length; i++) {
	    for (int j = 0; j < i; j++) {
		String big = null;
		String small = null;
		if (targetNames[i].startsWith(targetNames[j])) {
		    big = targetNames[i];
		    small = targetNames[j];
		} else if (targetNames[j].startsWith(targetNames[i])) {
		    big = targetNames[j];
		    small = targetNames[i];
		}
		if (big != null) {
		    String suffix = big.substring(small.length(),
						  big.length());
		    if (suffix.startsWith("" + CVC_PathSep())) {
			Set s = (Set) m.get(suffix);
			if (s == null) {
			    s = new HashSet();
			    m.put(suffix, s);
			}
			s.add(small);
		    }
		}
	    }
	}

	Set entries = m.entrySet();
	iter = entries.iterator();
	while (iter.hasNext()) {
	    Map.Entry e = (Map.Entry) iter.next();
	    Set prefixes = (Set) e.getValue();
	    if (prefixes.size() > 1) {
		assertFieldSemantics((String) e.getKey(), prefixes, sb);
	    }
	}
    }

    protected void assertFieldSemantics(String field, Set targets,
					StringBuffer sb)
	throws IOException {
	String[] targs = (String[]) targets.toArray(new String[0]);
	for (int i = 0; i < targs.length; i++) {
	    for (int j = 0; j < i; j++) {
		String assertion =
		    assertion(implies(binaryOp(targs[i], targs[j], CVC_EQ),
				      binaryOp(targs[i] + field,
					       targs[j] + field, CVC_EQ)));
		writeCommand(assertion, sb);
	    }
	}
    }

	// get the subset of targets that have reference type
    protected Set referenceTargets(Set targets) {
	Set targetsCopy = new HashSet(targets);
	Iterator iter = targetsCopy.iterator();
	while(iter.hasNext()) {
	    Predicate target = (Predicate) iter.next(); //added
	    
	    if (!target.type().isReference()
		|| (target instanceof PredicateArrayAccess
		    && ((PredicateArrayAccess)target).type().isArray()))
		iter.remove();	    
	}
	return targetsCopy;
    }

    
	// add an axiom about integers, since CVC only includes
	// the theory of rationals.  this axiom is the same as the
	// first heuristic used in Simplify's simplex engine, discussed
	// in their tech report.
	// unfortunately, this axiom seems to make CVC Lite unsound!
//     protected void assertIntegerAxioms(StringBuffer sb) throws IOException {
// 	String var1 = "v";
// 	String var2 = "w";
// 	String[] vars = {var1, var2};
// 	String[] types = {CVC_real, CVC_real};
// 	    // v > w ==> v >= (w + 1)
// 	String pred = implies(binaryOp(var1, var2, CVC_GT),
// 			      binaryOp(var1,
// 				       binaryOp(var2, "1", CVC_ADD),
// 				       CVC_GE));
// 	String assertion = assertion(forall(vars, types, pred));
// 	writeCommand(assertion, sb);
// //	newLine(sb);
//     }

    
	// output an assertion that t1 is a subtype of t2
    protected void assertSubtype(Type t1, Type t2, Set targets,
				 StringBuffer sb)
	throws IOException {
	String id1 = cleanIdentifier(t1.toString());
	String id2 = cleanIdentifier(t2.toString());

	Iterator iter = targets.iterator();
	while (iter.hasNext()) {
		//PredicateTarget target = (PredicateTarget) iter.next();
		//String var = cleanIdentifier(target.name());
	    Predicate target = (Predicate) iter.next();
	    String var = pred2String(target);

	    String assertion = assertion(implies(instanceOf(var, id1),
						 instanceOf(var, id2)));
	    writeCommand(assertion, sb);
	}
    }

	// output an assertion that t1 and t2 are disjoint
    protected void assertDisjoint(Type t1, Type t2, Set targets,
				  StringBuffer sb)
	throws IOException {
	String id1 = cleanIdentifier(t1.toString());
	String id2 = cleanIdentifier(t2.toString());

	Iterator iter = targets.iterator();
	while (iter.hasNext()) {
		// PredicateTarget target = (PredicateTarget) iter.next(); removed
	    Predicate target = (Predicate) iter.next();
		//String var = cleanIdentifier(target.name()); removed
	    String var = pred2String(target);		
	    String assertion = assertion(not(and(instanceOf(var, id1),
						 instanceOf(var, id2))));
	    writeCommand(assertion, sb);
	}
    }

    protected String and(String p1, String p2) {
	return binaryOp(p1, p2, CVC_AND);
    }

    protected String or(String p1, String p2) {
	return binaryOp(p1, p2, CVC_OR);
    }

    protected String not(String p) {
	return unaryOp(p, CVC_NOT);
    }

    protected String implies(String p1, String p2) {
	return binaryOp(p1, p2, CVC_IMP);
    }
    
    protected String iff(String p1, String p2) {
	return binaryOp(p1, p2, CVC_IFF);
    }    

    protected String binaryOp(String p1, String p2, String op) {
	return "(" + p1 + " " + op + " " + p2 + ")";
    }
    
    protected String unaryOp(String p, String op) {
	return op + "(" + p + ")";
    }
    
    protected String forall(String[] vars, String[] types, String pred) {
	    //assert(vars.length == types.length);
	String[] args = new String[vars.length];
	for (int i = 0; i < vars.length; i++) {
	    args[i] = decl(vars[i], types[i]);
	}
	String argStr = "(" + strList(args, ", ") + ")";
	return "FORALL " + argStr + ": " + pred;
    }

    protected String instanceOf(String specializee, String specializer) {
	String[] args = {specializee, specializer};
	return funCall(CVC_Instanceof, args);
    }

    protected String assertion(String s) {
	return "ASSERT " + s;
    }

    protected String query(String s) {
	return "QUERY " + s;
    }

    protected String push() {
	return "PUSH";
    }

    protected String pop() {
	return "POP";
    }
    
    protected String funCall(String funName, String[] args) {
        String r = funName;
        if (args.length > 0)
            r += "(" + strList(args, ", ") + ")";
        return r;
    }
	
    
    protected String typeDecl(String s) {
	return decl(s, "TYPE");
    }

    protected String predDecl(String name, String[] argTypes) {
	return funDecl(name, argTypes, CVC_bool);
    }

    protected String funDecl(String name, String[] argTypes,
			     String resType) {
        String t;
        if (argTypes.length == 0)
            t = resType;
        else {
            if (argTypes.length == 1)
                t = argTypes[0];
            else
                t = "[" + strList(argTypes, ", ") + "]";
            t = "[" + t + " -> " + resType + " ]";
            
        }
        return decl(name, t);
    }

    protected String decl(String id, String type) {
	return id + " : " + type;
    }

	// CVC can't deal with '.' or "[]" in identifier names 
    protected String cleanIdentifier(String id) {
	id = id.replace(".", CVC_PathSep());
	id = id.replace("[]", CVC_ArraySyntax());
 	return id;
    }

	// concatenate an array of strings, with separator sep
    protected String strList(String[] strs, String sep) {
	String res = "";
	int len = strs.length - 1;
	for (int i = 0; i < len; i++) {
	    res += strs[i] + sep;
	}
	if (strs.length > 0) {
	    res += strs[len];
	}
	return res;
    }

/** Ack! This doesn't work.  At least, it works in theory, but not in
 * practice, because of free variables in the query.  They are implicitly
 * universally quantified (and with an implicand that asserts their axioms).
 * So just negating the given query does not produce its logical negation.
 * You'd really need to existentially quantify over free variables.  Maybe
 * I'll think of doing that some day.
 **/
    
// 	// query for something and its negation.
// 	// this is a good sanity check, to make sure that we aren't
// 	// hitting CVC's incompleteness.  if CVC says Valid, then we
// 	// know the thing is true, but if it says InValid, then CVC may
// 	// just not be smart enough.  so if we get at least one Valid, we're
// 	// safe, and if we get two InValids, then we have found an
// 	// incompleteness.  (if we get two Valids, then CVC is unsound!)
//     protected void writeQueryAndNegation(String s, StringBuffer sb)
// 	throws IOException {
// 	writeQuery(s, sb);
// 	writeQuery(not(s), sb);
//     }
	
    
    protected void writeQuery(String s, StringBuffer sb) throws IOException {
//	writeCommand(push(), sb);
	writeCommand(query(s), sb);
//	writeCommand(pop(), sb);
	newLine(sb);
    }
    
    protected void writeCommand(String s, StringBuffer sb) throws IOException {
	sb.append(s + "; ");
    }

    protected void newLine(StringBuffer sb) throws IOException {
	sb.append("" + '\n');
    }

    protected String pred2String(Predicate p) {
	return pred2String(p, false);
    }	
    
    protected String pred2String(Predicate p, boolean stemsFromArrAccess) {
	if (p instanceof PredicateArrayAccess) {
	    PredicateArrayAccess paa = (PredicateArrayAccess) p;
	    return pred2String(paa.array(), true)
		+ "["
		+ pred2String(paa.index(), false)
		+ "]";
	} else if (p instanceof Specializer) {
	    Specializer s = (Specializer) p;
	    return instanceOf(pred2String(s.target(), false),
			      cleanIdentifier(ts.erasure(s.type()).toString()));
	} else if (p instanceof TruePredicate) {
	    return CVC_true;
	} else if (p instanceof PredicateCall) {
	    PredicateCall pc = (PredicateCall) p;
	    List args = pc.arguments();
	    int numArgs = args.size();
	    String[] argStrings = new String[numArgs];
	    for (int i = 0; i < numArgs; i++)
		argStrings[i] = pred2String((Predicate)args.get(i), false);
	    return funCall(cleanIdentifier(pc.name()), argStrings);
	} else if (p instanceof PredicateSpecial) {
	    return ((PredicateSpecial) p).name();
	} else if (p instanceof PredicateTarget) {
	    PredicateTarget pt = (PredicateTarget) p;
	    return (stemsFromArrAccess && pt.type().isArray()
		    ? mangleArrayName(cleanIdentifier(pt.name()),
				      pt.type().toArray())
		    : cleanIdentifier(pt.name()));
	} else if (p instanceof PredicateLit) {
	    return literal2String(((PredicateLit) p).value());
	} else if (p instanceof PredicateBinary) {
	    PredicateBinary pb = (PredicateBinary) p;
	    return binaryOp(pred2String(pb.left(), false),
			    pred2String(pb.right(), false),
			    binaryOperator2String(pb.operator(),
						  pb.left().type().isBoolean()));
	} else if (p instanceof PredicateUnary) {
	    PredicateUnary pu = (PredicateUnary) p;
	    return unaryOp(pred2String(pu.base(), false),
			   unaryOperator2String(pu.operator()));
	} else {
	    throw new RuntimeException("Don't know how to convert predicate "
				       + p + " to a string for CVC");
	}
    }
    
    protected String type2String(Type t) {
	if (t.isArray()) {
	    return CVC_array + type2String(((ArrayType)t).base());
	}
        if (t.isReference()) {
	    return CVC_Object();
	} else if (t.isNumeric()) {
	    return CVC_real;
	} else if (t.isBoolean()) {
	    return CVC_bool;
	} else {
	    throw new RuntimeException("Don't know how to translate type "
				       + t + " to CVC Lite.");
	}	
    }

    protected String binaryOperator2String(Binary.Operator op,
					   boolean operandsAreBoolean) {
	if (op.equals(Binary.COND_AND))
	    return CVC_AND;
	else if (op.equals(Binary.COND_OR))
	    return CVC_OR;
	else if (op.equals(Binary.GT))
	    return CVC_GT;
	else if (op.equals(Binary.GE))
	    return CVC_GE;
	else if (op.equals(Binary.LT))
	    return CVC_LT;
	else if (op.equals(Binary.LE))
	    return CVC_LE;
	else if (op.equals(Binary.EQ)) {
	    if (operandsAreBoolean)
		return CVC_IFF;
	    else
		return CVC_EQ;
	}
	else if (op.equals(Binary.NE)) {
	    if (operandsAreBoolean)
		return CVC_XOR;
	    else
		return CVC_NE;
	}
	else if (op.equals(Binary.ADD))
	    return CVC_ADD;
	else if (op.equals(Binary.SUB))
	    return CVC_SUB;
	else if (op.equals(Binary.MUL))
	    return CVC_MUL;
	else
	    throw new RuntimeException("Don't know how to convert " + op +
				       " to a CVC Lite operator");
    }

    protected String unaryOperator2String(Unary.Operator op) {
	if (op.equals(Unary.NOT))
	    return CVC_NOT;
	else if (op.equals(Unary.NEG))
	    return CVC_NEG;
	else 
	    throw new RuntimeException("Don't know how to convert " + op +
				       " to a CVC Lite operator");
    }

	
    protected String literal2String(Object literal) {
	if (literal instanceof Integer)
	    return literal.toString();
	else if (literal instanceof Boolean) {
		// CVC's booleans are TRUE and FALSE
	    return literal.toString().toUpperCase();
	} else if (literal == null) {
		// this is the null literal
	    return CVC_Null;
	}
	throw new RuntimeException("Don't know how to convert " + literal +
				   " to a CVC Lite literal");
    }

    protected void debugWrite(String s) throws IOException {
        if (DEBUG)
          fw.write(s);
    }

    protected void debugFlush() throws IOException {
        if (DEBUG)
          fw.flush();
    }

	// run cvc with the given input, and return the output string
	// (Valid. or Invalid.)
    protected boolean cvcQuery(StringBuffer sb) throws IOException {
	bw.write(push() + "; ");
	debugWrite(push() + "; ");
	bw.write(sb.toString() + '\n');
	debugWrite(sb.toString() + '\n');
	bw.flush();
	debugFlush();
	bw.write(pop() + "; ");
	debugWrite(pop() + "; ");
	int i = 0;
	while (i == 0) {
	    String line = br.readLine();
	    i = interpretCVCOutput(line);
	}
	if (i > 0)
	    return true;
	else
	    return false;
    }

// 	// parse back the results from querying a predicate and its
// 	// negation
//     protected boolean interpretQueryAndNegationOutput(String res1,
// 						      String res2) {
// 	boolean b1 = interpretCVCOutput(res1);
// 	boolean b2 = interpretCVCOutput(res2);
// 	if (b1 && b2) {
// 	    throw new RuntimeException("Found an unsoundness in CVC!");
// 	} else if (!b1 && !b2) {
// 	    throw new RuntimeException("Found an incompleteness in CVC!");
// 	} else {
// 	    return b1;
// 	}
//     }

    protected int interpretCVCOutput(String s) {	
	if (s.endsWith("InValid.")) {		
	    return -1;
	} else if (s.endsWith("Valid.")) {
	    return 1;
	} else if (s == null) {	    
	    throw new RuntimeException("No response from CVCL.  It does not understand"
				       + " the input.");
	} else {
		// we're reading a line of output that is not the
		// "Valid/Invalid line, e.g. part of a counterexample		
	    return 0;
	}
    }

    protected String executableName() {
	return "cvcl";
    }

	// the name of the CVC executable depends on the OS;
	// windows doesn't work without the .bat extension
    protected void determineCVCExecutable() {
	cvc_executable = executableName();
	
	    // check for the OS
	String osName = System.getProperty("os.name");
	if (osName.toLowerCase().startsWith("windows")) 
	    cvc_executable += ".exe";
    }
}
