package polyglot.ext.predj.types;

import java.util.*;

import polyglot.types.*;

/** The type information for a PredJDispatcher. **/
public class PredJDispatcherInstance_c implements PredJDispatcherInstance {

	// the list of mates (PredJMethodInstances)
    protected List dispatcherMates;

	// does the associated dispatcher include an abstract method?
	// note that the abstract method will not be among the method
	// instances in this.dispatcherMates.
    protected boolean hasAbstractMethod;

	// this is used to represent the partial order among the dispatcher
	// mates of this dispatcher.  each entry of the array contains a
	// set of the mate numbers (Integers, because Java Sets require things
	// to be boxed) of mates that are (transitively, but not reflexively)
	// overridden by the mate whose number is the entry's index.
    protected List[] partialOrder;

    	// for deserializing types
    protected PredJDispatcherInstance_c() {
    }
    
    public PredJDispatcherInstance_c(List mates,
				     boolean hasAbstractMethod,
				     PredJTypeSystem ts) {
	dispatcherMates = mates;
	this.hasAbstractMethod = hasAbstractMethod;

	partialOrder = new LinkedList[dispatcherMates.size()];
	for (int i = 0; i < partialOrder.length; i++) {
	    partialOrder[i] = new LinkedList();
	}
//	computeMatesPartialOrder(ts);
    }

    public List dispatcherMates() {
	return dispatcherMates;
    }

    public boolean hasAbstractMethod() {
	return hasAbstractMethod;
    }
    
    public List[] partialOrder() {
	return partialOrder;
    }


	// check that this dispatcher's gf is complete for the given
	// class type.
    public boolean isComplete(ClassType ct, PredJTypeSystem ts) {
	List mates = dispatcherMates();

	    // get method instances for all the mates and for all
	    // methods of this gf in superclasses
	PredJMethodInstance mi = (PredJMethodInstance) mates.get(0);
	List allMIs = ts.accessibleConcreteMethods(ct, mi);
	
	    // if there are no methods to consider, then we're certainly
	    // not complete
	if (allMIs.isEmpty()) {
	    return false;
	}
	
	    // if there is a concrete method that does not have a
	    // predicate, then it is a (local) default, so
	    // completeness is assured
	Iterator i = allMIs.iterator();
	boolean complete = false;
	while (i.hasNext()) {
	    PredJMethodInstance pmi = (PredJMethodInstance) i.next();
	    if (! pmi.hasPredicate()) {
		complete = true;
		break;
	    }
	}
	
	    // there is no default, so we have to use the theorem
	    // prover...
	return complete || ts.isCompletelyImplemented(allMIs);
    }
    
	// check this group of mates for unambiguity, throwing a semantic
	// exception if one is found.  in the process, we also compute
	// the mates partial order, for use during code generation.
    public void checkUnambiguity(PredJTypeSystem ts)
	throws SemanticException {
	
	    // make an array of the mates' method instances
	PredJMethodInstance[] mis =
	    (PredJMethodInstance[]) dispatcherMates.toArray(
		new PredJMethodInstance[0]);

	    // a list of pairs of mates that need a resolving set.
	    // pairs are represented by consecutive elements of the vector.
	Vector needResolving = new Vector();
	
	    // compare each pair of mates for ambiguity.  no need to
	    // compare a mate with an inherited method of a
	    // superclass, because of our encapsulated dispatch
	    // semantics, which implies that the mate always overrides
	    // such a method.
	for(int k = 0; k < mis.length; k++) {
	    PredJMethodInstance mi1 = mis[k];
	    for(int j = 0; j < k; j++) {
		PredJMethodInstance mi2 = mis[j];
		
		    // if the methods are disjoint, they're certainly
		    // unambiguous
		if (ts.areDisjoint(mi1, mi2)) {
		    continue;
		}
		boolean ksubj = ts.overrides(mi1, mi2);
		boolean jsubk = ts.overrides(mi2, mi1);
		    // if the methods override each other, they're
		    // equivalent and hence ambiguous.
		if (ksubj && jsubk) {
		    throw new SemanticException(
			"This method and the one at " +
			mi1.position().nameAndLineString()
			+ " have logically equivalent predicates.",
			mi2.position());
		}

		    // if one is strictly more specific than the
		    // other, then the methods are unambiguous.  we
		    // also record their relationship in the partial
		    // order data structure.
		if (ksubj) {
		    partialOrder[k].add(new Integer(j));
		    continue;
		}
		if (jsubk) {
		    partialOrder[j].add(new Integer(k));
		    continue;
		}
		
		    // if we get here, then the methods intersect, and
		    // we need to check for an appropriate set of
		    // resolving methods.  we can't do that, however,
		    // until the partial order has been finished being
		    // built.  so for now we just record that the
		    // methods need resolving.
		needResolving.add(new Integer(k));
		needResolving.add(new Integer(j));
	    }
	}

	    // now that we've built the partial order, check for resolving
	    // methods.
	Iterator iter = needResolving.iterator();
	while (iter.hasNext()) {
	    int k = ((Integer) iter.next()).intValue();
	    int j = ((Integer) iter.next()).intValue();
	    
	    if (!ts.haveResolvingMethods(k, j, partialOrder, mis)) {
		    // we've found an ambiguity		
		throw new SemanticException(
		    "This method is ambiguous with the one at " +
		    mis[k].position().nameAndLineString()
		    + ".", mis[j].position());
	    }
	    
	}
    }
    
	// check that each mate's predicate can be satisfied, throwing a
	// semantic exception if one is not.
    public void checkSatisfiablePredicates(PredJTypeSystem ts)
	throws SemanticException
    {
	    // make an array of the mates' method instances
	PredJMethodInstance[] mis =
	    (PredJMethodInstance[]) dispatcherMates.toArray(
		new PredJMethodInstance[0]);

	    // check that each method is reachable
	for(int k = 0; k < mis.length; k++) {
	    PredJMethodInstance mi = mis[k];

	    if (!ts.satisfiablePredicate(mi)) {
		throw new SemanticException(
		    "This method's predicate is not satisfiable,"
		    + " making the method unreachable.", mi.position());
	    }
	}
    }

// 	// check this group of mates for unambiguity, throwing a semantic
// 	// exception if one is found.
//     public void checkUnambiguity(PredJTypeSystem ts)
// 	throws SemanticException {
	
// 	    // make an array of the mates' method instances
// 	PredJMethodInstance[] mis =
// 	    (PredJMethodInstance[]) dispatcherMates.toArray(
// 		new PredJMethodInstance[0]);

// 	List[] po = partialOrder();
	
// 	    // compare each pair of mates for ambiguity.  no need to
// 	    // compare a mate with an inherited method of a
// 	    // superclass, because of our encapsulated dispatch
// 	    // semantics, which implies that the mate always overrides
// 	    // such a method.
// 	boolean equivalent;
// 	boolean unambiguous;
// 	for(int k = 0; k < mis.length; k++) {
// 	    for(int j = 0; j < k; j++) {
// 		PredJMethodInstance mi1 = mis[k];
// 		PredJMethodInstance mi2 = mis[j];
// 		    // here we rely on the fact that the mate number of a
// 		    // mate is simply based on the order in which it appears
// 		    // in the PredJDispatcher's list of dispatchers.  so
// 		    // mi1 has mate number k, and mi2 has mate number j.
// 		boolean ksubj = po[k].contains(new Integer(j));
// 		boolean jsubk = po[j].contains(new Integer(k));
// 		    // if the two method predicates are equivalent, then
// 		    // we have an ambiguity
// 		equivalent = ksubj && jsubk;
// 		if (!equivalent) {

// 			// otherwise if one method overrides the other, there
// 			// is no ambiguity
// 		    unambiguous = ksubj || jsubk;
		    
			    
// 			// otherwise we need to do some theorem proving!

// 			// are the two methods disjoint?
// 		    unambiguous = unambiguous || ts.areDisjoint(mi1, mi2);

// 			// if not, do the methods that override both of
// 			// these methods cover their intersection?
// 		    unambiguous = unambiguous ||
// 			ts.haveResolvingMethods(k, j, partialOrder, mis);
// 		}
		
// 		if (equivalent || !unambiguous) {
// 			// we've found an ambiguity	
// 		    throw new SemanticException(
// 			"This method is ambiguous with the one at " +
// 			mi2.position().nameAndLineString()
// 			+ ".", mi1.position());
// 		}
// 	    }
// 	}
//     }
    

//      protected void computeMatesPartialOrder(PredJTypeSystem ts) {
//  	PredJMethodInstance[] mates =
//  	    (PredJMethodInstance[]) dispatcherMates.toArray(
//  		new PredJMethodInstance[0]);

// 	    // if revMap[i] includes j, then mate j subtypes mate i
// 	List[] revMap = new List[mates.length];
// 	for (int i = 0; i < revMap.length; i++) {
// 	    revMap[i] = new LinkedList();
// 	}
	
// 	    // record whether we've tested a given pair of mates already
// 	boolean[][] todo = new boolean[mates.length][mates.length];
// 	for (int i = 0; i < mates.length; i++) {
// 	    for (int j = 0; j < mates.length; j++) {
// 		if (i == j)
// 			// we don't need to test a mate against itself
// 		    todo[i][j] = false;
// 		else 
// 		    todo[i][j] = true;
// 	    }
// 	}

//  	for(int i=0; i < mates.length; i++) {
//  	    for (int j=0; j < mates.length; j++) {
//  		if (todo[i][j]) {
//  		    if (ts.overrides(mates[i], mates[j])) {
// 			propagateSub(i, j, revMap, todo);
//  		    } else {
// 			propagateNoSub(i, j, revMap, todo);
// 		    }
//  		}
//  	    }
//  	}
//      }


//     void propagateSub(int i, int j, List[] revMap, boolean[][] todo) {
// 	todo[i][j] = false;
// 	partialOrder[i].add(new Integer(j));
// 	revMap[j].add(new Integer(i));
// 	Iterator iter = partialOrder[j].iterator();
// 	while (iter.hasNext()) {
// 	    Integer v = (Integer) iter.next();
// 	    int val = v.intValue();
// 	    if (todo[i][val]) {
// 		propagateSub(i, val, revMap, todo);
// 	    }
// 	}
// 	iter = revMap[i].iterator();
// 	while (iter.hasNext()) {
// 	    Integer v = (Integer) iter.next();
// 	    int val = v.intValue();
// 	    if (todo[val][j]) {
// 		propagateSub(val, j, revMap, todo);
// 	    }
// 	}
//     }


//     void propagateNoSub(int i, int j, List[] revMap, boolean[][] todo) {
// 	todo[i][j] = false;
// 	Iterator iter = partialOrder[i].iterator();
// 	while (iter.hasNext()) {
// 	    Integer v = (Integer) iter.next();
// 	    int val = v.intValue();
// 	    if (todo[val][j]) {
// 		propagateNoSub(val, j, revMap, todo);
// 	    }
// 	}

// 	iter = revMap[j].iterator();
// 	while (iter.hasNext()) {
// 	    Integer v = (Integer) iter.next();
// 	    int val = v.intValue();
// 	    if (todo[i][val]) {
// 		propagateNoSub(i, val, revMap, todo);
// 	    }
// 	}
//     }


}
