package polyglot.ext.predj.util;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import java.util.*;

// Make sure that an expression is a legal predicate expression.
// If not, throw a SemanticException.
// Current Restrictions Include:
//  *) Check for liearity if we have multiplication.
//  *) Check for legal array accesses.
//     - An array access may not access a field.
//     - An array access that is specialized on may not have array type.
//  *) Check for legal usage of method calls.
//     - The result of a method call may not access a field.
//     - The result of a method call that is specialized on may not have
//       array type.

public class checkPredicateRestrictions {

    public static void checkPredicateRestrictions(Expr e)
	throws SemanticException {
	if (e instanceof ArrayAccess) {
	    ArrayAccess a = (ArrayAccess) e;
	    checkPredicateRestrictions(a.array());
	    checkPredicateRestrictions(a.index());	
	} else if ((e instanceof Local) || (e instanceof Lit) ||
		   (e instanceof BindingNode) || (e instanceof Special)) {
		// Take no action.
	} else if (e instanceof Cast) {
	    Cast cast = (Cast) e;
	    checkPredicateRestrictions(cast.expr());
	} else if (e instanceof Call) {
	    Call c = (Call) e;
	    List args = c.arguments();
	
	    checkPredicateRestrictions((Expr) c.target());
	    Iterator i = args.iterator();
	    while (i.hasNext())
		checkPredicateRestrictions((Expr) i.next());	
	} else if (e instanceof Unary) {
	    Unary u = (Unary) e;
	    checkPredicateRestrictions(u.expr());
	} else if (e instanceof Binary) {
	    Binary b = (Binary) e;
	    Expr left = (Expr)b.left();
	    Expr right = (Expr)b.right();

	    // Check for Linearity.
	    // if this is a multiplication, one of the two subexprs
	    // must be a constant
	    if (b.operator() == Binary.MUL) {
		if (! (left.isConstant() || right.isConstant())) {
		    throw new SemanticException("All arithmetic expressions in "
						+ "predicates must be linear.",
						b.position());
		}
	    }
	    checkPredicateRestrictions(left);
	    checkPredicateRestrictions(right);
	} else if (e instanceof Field) {
	    Field f = (Field) e;
	    if (f.target() instanceof ArrayAccess)
		throw new SemanticException("Currently, Array Accesses in predicates "
					    + "may not access a field from the result "
					    + "of the array access.", f.position());
	    else if (f.target() instanceof Call)
		throw new SemanticException("Currently, Method Calls in predicates "
					    + "may not access a field from the result "
					    + "of the call.", f.position());
	    else if (!(f.target() instanceof Expr)) {
		    // This is of Receiver type, but not an Expr.
		    // No action needed.
	    }
	    else
		checkPredicateRestrictions((Expr)f.target());
	} else if (e instanceof AbstractSpecializerNode) {
	    AbstractSpecializerNode sn = (AbstractSpecializerNode) e;
	    if ((sn.target() instanceof ArrayAccess) &&
		sn.target().type().isArray())
		throw new SemanticException("A specializer in a predicate may not be both "
					    + "an Array Access and still of array type.",
					    sn.position());
	    else if((sn.target() instanceof Call) &&
		    sn.target().type().isArray()) {
		throw new SemanticException("A specializer in a predicate may not be both "
					    + "a method call and still of array type.",
					    sn.position());	    
	    }	
	    else
		checkPredicateRestrictions(sn.target());
	} else {
	    throw new RuntimeException(
		"checkPredicateRestrictions called on an unexpected expression " + e);
	}
    }
}
