package polyglot.ext.predj.visit;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.jl5.types.*;
import polyglot.util.*;

import polyglot.frontend.Job;

import polyglot.ext.predj.util.toJavaExpr;

import java.util.*;

/** Visitor that translates PredJ AST nodes to Java AST nodes, for
    subsequent output. **/
public class PredJTranslator extends ContextVisitor {

	// String for mangling the name of each dispatchee method
    protected static String dispatcheeStr = "$body";

	// String for making the formal names of the dispatcher method
    public static String argStr = "arg$";
    
    public PredJTranslator(Job job, TypeSystem ts, NodeFactory nf) {
	super(job, ts, nf);
    }

    protected static String dispatcheeName(String name, int mateNumber) {
	return name + dispatcheeStr + mateNumber;
    }

    protected static String type2String(Type t) {
	if (t.isReference()) {
	    return "Object";
	} else if (t.isNumeric()) {
	    return "Real";
	} else if (t.isBoolean()) {
	    return "Boolean";
	} else {
	    throw new RuntimeException("Unexpected type " + t +
		" in a predicate.");
	}	
    }
	// give each argument a unique name based on its position.
	// also use the static type as part of the name, because CVC will
	// only declare a given name once, and without the type we'd need
	// the same argname to have different types for different queries.
    public static String argName(int formalNum, Type t) {
	return argStr + formalNum + "$" + type2String(t);
    }

    protected Node clearPureFlag(MethodDecl md) {
      return md.flags(md.flags().clear(((PredJTypeSystem)typeSystem()).Pure()));
    }

    protected Node leaveCall(Node n) throws SemanticException {
	if (n instanceof PredJPredMethodDecl) {
	    PredJPredMethodDecl node =
		(PredJPredMethodDecl) leaveMethod((PredJPredMethodDecl) n);

		// we need to update formals to have their specialized
		// types, if they have them
	    ArrayList formals = new ArrayList(node.formals());
	    Iterator iter = node.formalBindings().iterator();
	    while (iter.hasNext()) {
		SpecializerNode sn = (SpecializerNode) iter.next();
		int paramPos = sn.paramPosition();
		// update the paramPos'th formal
		PredJFormal f = (PredJFormal) formals.get(paramPos);
		formals.set(paramPos, f.type(sn.specializer()));
	    }

	    // we also need to add any new variable bindings from the
	    // predicate to the body
	    Block body = node.body();
	    BindingNode[] bindings =
		(BindingNode[])
		node.localBindings().toArray(new BindingNode[0]);
	    // prepend the bindings to the body in reverse order, so
	    // that they will be executed in the forward order
	    for (int i = bindings.length - 1; i >=0; i--) {
		BindingNode bn = bindings[i];
		body = body.prepend(bn.localBinding());
	    }
 		
	    return clearPureFlag((MethodDecl)node.formals(formals).body(body));
	} else if (n instanceof PredJMethodDecl) {
	    return leaveMethod((PredJMethodDecl) n);
	} else if (n instanceof PredJDispatcher) {
	    PredJDispatcher pd = (PredJDispatcher) n;
		// find some linear ordering of the dispatcher mates
		// that is consistent with the partial order; i.e.  a
		// topological sort the linearization is represented
		// as an array of the dispatcher mate numbers (ints),
		// from most- to least-specific	    
	    int[] mateNums = linearize(pd.partialOrder());

		// create the dispatcher method
	    List mates = pd.dispatcherMates();
	
	    PredJMethodDecl m = (PredJMethodDecl) mates.get(0);
	    // make the formals, and create accesses to these formals,
	    // which are used as actual args in calls from the dispatcher
	    // method
	    LinkedList formals = new LinkedList();
	    LinkedList actualArgs = new LinkedList();
	    Iterator iter = m.formals().iterator();
	    int count = 0;
	    while (iter.hasNext()) {
		Formal f = (Formal) iter.next();
		String newName = argName(count, f.type().type());
		Formal newF = f.name(newName);
		formals.add(newF);
		actualArgs.add(nf.Local(null, newName));
		count++;
	    }	
	    // make the body
	    boolean returnTypeIsVoid = m.returnType().type().isVoid();
	    Stmt bodyStmt =
		createDispatcherMethodBody(mateNums, mates, 0, m.name(),
					   actualArgs, returnTypeIsVoid);
	    Block body = nf.Block(null, bodyStmt);
	
	    PredJMethodDecl dispMethod =
		(PredJMethodDecl) ((PredJNodeFactory)nf).JL5MethodDecl(
		    null, new FlagAnnotations(m.flags(), m.annotations()),
		    m.returnType(), m.name(), formals,
		    m.throwTypes(), body, m.paramTypes());

	    // need to give the method decl a method instance, since
	    // it would have been constructed on an earlier pass (and is
	    // needed on a later pass)
	    dispMethod = (PredJMethodDecl) dispMethod.methodInstance(
		dispMethod.makeMethodInstance(
		    context().currentClassScope(), ts));

	    return clearPureFlag(dispMethod);
	} else if (n instanceof ClassBody) {
	    ClassBody cb = (ClassBody) n;
	    // get rid of any abstract methods that have dispatcher
	    // mates.  they will not appear in the Java translation, since
	    // they have concrete mates who must be implemented there.
	    List members = new ArrayList(cb.members());
	    Iterator i = members.iterator();
	    while (i.hasNext()) {
		ClassMember cm = (ClassMember) i.next();
		if (cm instanceof PredJMethodDecl) {
		    PredJMethodDecl md = (PredJMethodDecl) cm;
		    if (md.flags().isAbstract() && md.hasMates()) {
			i.remove();
		    }
		}
	    }
	    return cb.members(members);
	} else
	    return super.leaveCall(n);
    }

    protected Node leaveMethod(PredJMethodDecl n)
	throws SemanticException {
//	PredJMethodInstance mi = (PredJMethodInstance) n.methodInstance();
	if (n.hasMates()) {
		// give this method a unique name
	    String name = dispatcheeName(n.name(), n.mateNumber());
	    n = (PredJMethodDecl) n.name(name);
		// method instances are no longer used, so they need
		// not be kept in sync
//	    mi = (PredJMethodInstance) mi.name(name);
		// make the method private
	    Flags flags = n.flags();
	    flags = flags.clearPublic().clearProtected();
	    flags = flags.Private();
	    n = (PredJMethodDecl) n.flags(flags);
//	    mi = (PredJMethodInstance) mi.flags(flags);
//	    n = n.methodInstance(mi);

	}
	
        return clearPureFlag(n);
    }

    protected Stmt createDispatcherMethodBody(int[] mateNums, List mates,
					      int index, String origMethodName,
					      List actualArgs,
					      boolean returnTypeIsVoid) {
	int currMateNum = mateNums[index];
	PredJMethodDecl currMate =
	    (PredJMethodDecl) mates.get(currMateNum);

	boolean hasPred = currMate.hasPredicate();

	    // create the call to the mate at the current index
	List castActuals = new LinkedList(actualArgs);
	PredJPredMethodDecl predMate = null;
	if (hasPred) {
	    predMate = (PredJPredMethodDecl) currMate;
		// cast actual args to their specialized types before
		// invoking the predicate method
	    Iterator i = predMate.formalBindings().iterator();
	    while (i.hasNext()) {
		SpecializerNode sn = (SpecializerNode) i.next();
		int paramPos = sn.paramPosition();		
		Expr actualArg = (Expr) castActuals.get(paramPos);
		castActuals.set(paramPos,
				nf.Cast(null, sn.specializer(), actualArg));
	    }
	}
	Stmt callStmt =
	    createCallToMate(currMateNum, origMethodName,
			     castActuals, returnTypeIsVoid);

	Stmt bodyStmt;

	    // if this method has a predicate, we need to do a
	    // conditional test for applicability, and then
	    // generate the rest of the dispatcher method
	    // as the else branch.  however, if this is the last mate to
	    // be generated and there are no methods in any superclasses
	    // to try, then no applicability test is necessary, because
	    // this method must be applicable (assuming prior completeness
	    // checking).  MultiJava requires default cases, so this scenario
	    // could never arise there.  But we don't always require them --
	    // things can be complete without them (e.g. (when P), (when !P)).
	boolean isLastMate = index + 1 == mateNums.length;
	if (hasPred &&
	    (!isLastMate || gfHasInheritedMethods(currMate))) {
		// create the predicate test
	    Expr testExpr =
		(Expr) toJavaExpr.toJavaExpr(
		    predMate.predNode().predicateExpr(), nf,
		    (PredJTypeSystem) ts);
	    
		// create the alternative if this method is not applicable
	    Stmt elseStmt;
	    if (isLastMate) {
		    // this is the last mate, so we should generate a
		    // super call
		elseStmt = createSuperCall(origMethodName, actualArgs,
					   returnTypeIsVoid);
	    } else {
		    // this is not the last mate, so recursively generate
		    // the rest of the dispatcher method body	
		index++;
		elseStmt =
		    createDispatcherMethodBody(mateNums, mates, index,
					       origMethodName, actualArgs,
					       returnTypeIsVoid);
	    }
	    
	    bodyStmt = nf.If(null, testExpr, callStmt, elseStmt);
	    
	} else {
		// this method is definitely applicable.  therefore,
		// just call it.
	    bodyStmt = callStmt;
	}
	
	return bodyStmt;
    }
    
    protected Stmt createCallToMate(int mateNum, String origMethodName,
				    List actualArgs,
				    boolean returnTypeIsVoid) {
	return createCallAndReturn(null,
				   dispatcheeName(origMethodName, mateNum),
				   actualArgs, returnTypeIsVoid);
    }
    
    protected Stmt createSuperCall(String methodName, List actualArgs,
				   boolean returnTypeIsVoid) {
	return createCallAndReturn(nf.Super(null), methodName,
				   actualArgs, returnTypeIsVoid);
    }

	// create a call to receiver.methodName, where the actuals are
	// simple local variables (actualArgs -- a list of Locals).
	// if the call returns a value, return that value also as the
	// result of the caller.
    protected Stmt createCallAndReturn(Receiver receiver, String methodName,
				       List actualArgs,
				       boolean returnTypeIsVoid) {
	Expr call = nf.Call(null, receiver, methodName, actualArgs);
	if (returnTypeIsVoid) {
	    return nf.Eval(null, call);
	} else {
	    return nf.Return(null, call);
	}
    }
    
    protected int[] linearize(List[] partialOrder) {
	int n = partialOrder.length;

	boolean[] marks = new boolean[n];
	for(int i = 0; i < n; i++) {
	    marks[i] = false;
	}

	int[] result = new int[n];
	int index = n-1;
	for(int i = 0; i < n; i++) {
	    index = dfs(partialOrder, result, marks, i, index);
	}

	return result;
    }

	// a variant of depth-first search, the heart of the
	// topological sort
    protected int dfs(List[] partialOrder, int[] result, boolean[] marks,
		      int i, int index) {
	if (!marks[i]) {
	    marks[i] = true;
		// recursively visit all the overridden mates
	    Iterator mates = partialOrder[i].iterator();
	    while (mates.hasNext()) {
		Integer mateNum = (Integer) mates.next();
		index = dfs(partialOrder, result, marks,
			    mateNum.intValue(), index);
	    }
	    result[index] = i;
	    index--;
	}
	return index;
    }

	// does the class containing this set of mates inherit at
	// least one concrete method of the same gf from a superclass?
    protected boolean gfHasInheritedMethods(MethodDecl currMate) {
	PredJMethodInstance currMI =
	    (PredJMethodInstance) currMate.methodInstance();
	ClassType currClass = (ClassType) currMI.container();
	PredJTypeSystem ts = (PredJTypeSystem) typeSystem();

	List allMIs = ts.accessibleConcreteMethods(currClass, currMI);

	Iterator i = allMIs.iterator();
	while (i.hasNext()) {
	    MethodInstance mi = (MethodInstance) i.next();
	    if (currClass != mi.container()) {
		return true;
	    }
	}
	return false;
    }
}

