package polyglot.ext.predj.visit;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.predj.util.canonicalize;
import polyglot.visit.*;

import polyglot.frontend.Job;


// canonicalize each predicate expression: substitute any new var
// bindings with their definitions, substitute a canonical name for
// each formal, cast formals and locals to their specialized
// types, and replace compile-time constant expressions by their values.

public class PredicateCanonicalizer extends ContextVisitor {

	// should we canonicalize the current node?  only if it's part of
	// a predicate.
    protected boolean shouldCanonicalize = false;
    
    public PredicateCanonicalizer(Job job, TypeSystem ts, NodeFactory nf) {
	super(job, ts, nf);
    }

    public PredicateCanonicalizer(PredicateCanonicalizer pc,
				  boolean shouldCanonicalize) {
	this(pc.job(), pc.typeSystem(), pc.nodeFactory());
	this.outer = pc.outer;
	this.context = pc.context;
	this.shouldCanonicalize = shouldCanonicalize;
    }
    
    protected NodeVisitor enterCall(Node n) throws SemanticException {
	if (n instanceof PredicateNode)
	    return new PredicateCanonicalizer(this, true);
	else if (n instanceof PredicateVarBinding) 
		// don't canonicalize a binding's local var decl, because
		// it is eventually used inside the method, rather than
		// generically as part of the gf.
	    return new PredicateCanonicalizer(this, false);
	else
	    return super.enterCall(n);
    }

    protected Node leaveCall(Node n) throws SemanticException {
	if (n instanceof Local) {
	    Local nl = (Local) n;
	    Expr newN = leaveExpr(nl);
	    if (shouldCanonicalize) {
		return canonicalize.canonicalize(newN, (PredJContext) context(),
						 nodeFactory());
	    }
	    return newN;
	} else if (n instanceof Expr) {
	    Expr ne = (Expr) n;
	    return leaveExpr(ne);
	} else
	    return super.leaveCall(n);
    }

    protected Expr leaveExpr(Expr n) {
	    // replace the expression by its constant value, if it has one.
	    // we only currently handle integer, boolean, and null literals.
	if (shouldCanonicalize && n.isConstant()) {
	    Expr newExpr = null;
	    Type type = n.type();
	    if (type.isBoolean()) {
		boolean val = ((Boolean) n.constantValue()).booleanValue();
		newExpr = nf.BooleanLit(null, val);
 	    } else if (type.isInt()) {
		int val = ((Integer) n.constantValue()).intValue();
 		newExpr = nf.IntLit(null, IntLit.INT, val);
	    } else if (type.isNull()) {
		newExpr = nf.NullLit(null);
	    }
	    if (newExpr != null) {
		return newExpr.type(type);
	    }
	}
	return n;
    }
}    
